#include "haversine.h"
#include "point_types.h"
#include "processor/proc_argparser.h"
#include "profiler/timer.h"
#include "json/dstring.h"
#include "json/json_entities.h"
#include "json/parser.h"
#include <aliases.h>
#include <assert.h>
#include <float.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

int main(int argc, char *argv[]) {
  u64 cpu_freq = get_cpu_freq(500);

  u64 cli_parse_start = read_cpu_timer();

  ProcessorArgs args = parse_args(argc, argv);

  u64 json_parse_start = read_cpu_timer();

  jentity_t *root = load_json(args.filepath);

  assert(root->type == JENTITY_SINGLE && root->value.type == JVAL_COLLECTION);

  u64 load_pairs_json_start = read_cpu_timer();

  jentity_t *pairs = root->value.collection->begin;

  assert(pairs->type == JENTITY_PAIR &&
         pairs->pair.value.type == JVAL_COLLECTION);

  u64 pair_count = pairs->pair.value.collection->size;

  PointPair *point_pairs = (PointPair *)malloc(sizeof(PointPair) * pair_count);
  memset(point_pairs, 0, pair_count);

  u64 index = 0;

  for (jentity_t *pair = pairs->pair.value.collection->begin; pair != NULL;
       pair = pair->next) {
    assert(index < pair_count && pair->type == JENTITY_SINGLE &&
           pair->value.type == JVAL_COLLECTION &&
           pair->value.collection->size == 4);

    jentity_t *x0 = pair->value.collection->begin;
    jentity_t *y0 = x0->next;
    jentity_t *x1 = y0->next;
    jentity_t *y1 = x1->next;

    PointPair p = {
        {x0->pair.value.num_dbl, y0->pair.value.num_dbl},
        {x1->pair.value.num_dbl, y1->pair.value.num_dbl},
    };

    point_pairs[index++] = p;
  }

  u64 binary_file_read_start = read_cpu_timer();

  const char *filename = "count_and_distances";

  FILE *fp = fopen(filename, "r");
  if (!fp) {
    printf("Failed to open the %s file", filename);
  } else {
    // Skip the count
    fseek(fp, sizeof(u64), SEEK_SET);
  }

  u64 haversine_sum_start = read_cpu_timer();

  f64 sum = 0.0;
  f64 distance = 0.0;
  f64 saved_distance = 0.0;
  for (u64 i = 0; i < pair_count; ++i) {
    distance = haversine_of_degrees(point_pairs[i], EARTH_RADIUS_KM);

    if (fp) {
      fread(&saved_distance, sizeof(f64), 1, fp);

      if (fabs(distance - saved_distance) > FLT_EPSILON) {
        printf("%llu: %.16f does not equal %.16f\n", (unsigned long long)i,
               distance, saved_distance);
      }
    }

    sum += distance;
  }

  u64 haversine_average_start = read_cpu_timer();

  printf("\nAVERAGE DISTANCE: %f\n", sum / pair_count);

  u64 tear_down_start = read_cpu_timer();

  if (fp) {
    fclose(fp);
  }

  free(point_pairs);

  u64 end = read_cpu_timer();

  if (cpu_freq) {
    u16 time_precision = 16;
    u16 time_char_count = 20;
    u16 percentage_precision = 8;
    u16 percentage_char_count = 12;

    f64 total = (f64)end - (f64)cli_parse_start;
    f64 cli_parse = (f64)json_parse_start - (f64)cli_parse_start;
    f64 json_parse = (f64)load_pairs_json_start - (f64)json_parse_start;
    f64 json_load = (f64)binary_file_read_start - (f64)load_pairs_json_start;
    f64 binary_open = (f64)haversine_sum_start - (f64)binary_file_read_start;
    f64 haversine_sum = (f64)haversine_average_start - (f64)haversine_sum_start;
    f64 haversine_average = (f64)tear_down_start - (f64)haversine_average_start;
    f64 tear_down = (f64)end - (f64)tear_down_start;

    printf("\n==========================PROFILING==========================\n");
    printf("Total: %*.*f seconds (CPU frequency: %llu)\n\n", time_char_count,
           time_precision, total / cpu_freq, (unsigned long long)cpu_freq);
    printf("  CLI PARSING: %*.*f seconds (%*.*f %%)\n", time_char_count,
           time_precision, cli_parse / cpu_freq, percentage_char_count,
           percentage_precision, cli_parse / total * 100.0);
    printf(" JSON PARSING: %*.*f seconds (%*.*f %%)\n", time_char_count,
           time_precision, json_parse / cpu_freq, percentage_char_count,
           percentage_precision, json_parse / total * 100.0);
    printf(" JSON LOADING: %*.*f seconds (%*.*f %%)\n", time_char_count,
           time_precision, json_load / cpu_freq, percentage_char_count,
           percentage_precision, json_load / total * 100.0);
    printf("  BINARY OPEN: %*.*f seconds (%*.*f %%)\n", time_char_count,
           time_precision, binary_open / cpu_freq, percentage_char_count,
           percentage_precision, binary_open / total * 100.0);
    printf("HAVERSINE SUM: %*.*f seconds (%*.*f %%)\n", time_char_count,
           time_precision, haversine_sum / cpu_freq, percentage_char_count,
           percentage_precision, haversine_sum / total * 100.0);
    printf("HAVERSINE AVG: %*.*f seconds (%*.*f %%)\n", time_char_count,
           time_precision, haversine_average / cpu_freq, percentage_char_count,
           percentage_precision, haversine_average / total * 100.0);
    printf("     TEARDOWN: %*.*f seconds (%*.*f %%)\n", time_char_count,
           time_precision, tear_down / cpu_freq, percentage_char_count,
           percentage_precision, tear_down / total * 100.0);
  }

  return 0;
}
